/*
Copyright (C) 2000,2001 Stefan Duffner 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <iostream>
#include <qregexp.h>

#include "ExportVHDL.h"
#include "Machine.h"
#include "TransitionInfo.h"
#include "IOInfo.h"
#include "IOInfoList.h"
#include "Convert.h"
#include "Utils.h"
#include "Options.h"

//using namespace std;


ExportVHDL::ExportVHDL(Options* opt)
  	  : Export(opt)
{
}


/// Writes all the relevant data into the tdf file.
void ExportVHDL::doExport()
{
  state_encoding=!options->getVHDLSymbolicStates();
  io_names=options->getVHDLInOutNames();
  synchronous_reset=options->getVHDLSyncReset();
  neg_reset=options->getVHDLNegReset();
  use_std_logic=options->getVHDLStdLogic();
  io_header=options->getVHDLIOheader();
  alliance=options->getVHDLAlliance();
  cond_notation=options->getVHDLCondNotation();

  writeHeader("--", "");
  if (io_header)
    writeDescription();
  writeEnvironment();
  writeEntity();
  writeArchitecture();
}


QString ExportVHDL::fileFilter()
{
  return "VHDL (*.vhdl *.vhd)";
}

/// Writes the 'library' and 'use' lines to the output stream
void ExportVHDL::writeEnvironment()
{
  using namespace std;

  if (use_std_logic)
  {
    *out << "LIBRARY IEEE;" << endl << endl;
    *out << "USE IEEE.std_logic_1164.ALL;" << endl << endl;
  }
}

/// Writes the 'entity' block to the output stream
void ExportVHDL::writeEntity()
{
  using namespace std;

  QString bit_string;
  bool io_names_t, state_encoding_t, synchronous_reset_t;

  if (alliance)
  {
    io_names_t=false;
    state_encoding_t=false;
    synchronous_reset_t=true;
  }
  else
  {
    io_names_t=io_names;
    state_encoding_t=state_encoding;
    synchronous_reset_t=synchronous_reset;
  }

  if (use_std_logic)
    bit_string = "std_ulogic";
  else
    bit_string = "bit";

  mname = machine->getName();
  mname.replace(QRegExp(" "), "_");

  *out << "ENTITY " << mname.latin1() << " IS" << endl;
  //*out << "--  generic (prop_delay : Time := 10 ns);" << endl;
  *out << "  PORT (clk: IN " << bit_string.latin1() << ";" << endl;
  if(synchronous_reset_t)
  {
    if (neg_reset)
      *out << "        srst_n: IN " << bit_string.latin1() << ";" << endl;
    else
      *out << "        srst_p: IN " << bit_string.latin1() << ";" << endl;
  }
  if (!alliance)
  {
    if (neg_reset)
      *out << "        rst_n: IN " << bit_string.latin1() << ";" << endl;
    else
      *out << "        rst_p: IN " << bit_string.latin1() << ";" << endl;
  }

  if (io_names_t)
  {
    QStringList inputs=machine->getInputNameList();
    QStringList::const_iterator i;
    for (i = inputs.constBegin(); i != inputs.constEnd(); ++i)
      *out << "        " << i->latin1() << ": IN "<< bit_string.latin1() << ";\n";

    if (machine->getNumMooreOutputs()>0)
    {
      QStringList mooreOutputs=machine->getMooreOutputList();
      for (i = mooreOutputs.constBegin(); i != mooreOutputs.constEnd(); ++i)
      {
	*out << "        " << i->latin1() << ": OUT "<< bit_string.latin1();
	if((i+1)!=mooreOutputs.constEnd() || machine->getNumOutputs()>0)
	  *out<< ";\n";
      }
    }
    if (machine->getNumOutputs()>0)
    {
      QStringList mealyOutputs=machine->getOutputNameList();
      for (i = mealyOutputs.constBegin(); i != mealyOutputs.constEnd(); ++i)
      {
	*out << "        " << i->latin1() << ": OUT "<< bit_string.latin1();
	if((i+1)!=mealyOutputs.constEnd())
	  *out<< ";\n";
      }
    }
    *out << ");\n";
  }
  else // ! io_names
  {
    *out << "        a: IN " << bit_string.latin1() << "_vector(" << machine->getNumInputs()-1 
      << " DOWNTO 0)";
    if (machine->getNumMooreOutputs()>0)
    {
      *out << ";\n        q: OUT " << bit_string.latin1() << "_vector(" << machine->getNumMooreOutputs()-1
	<< " DOWNTO 0)";
    }
    if (machine->getNumOutputs()>0)
    {
      *out << ";\n        o: OUT " << bit_string.latin1() << "_vector(" << machine->getNumOutputs()-1
	<< " DOWNTO 0)";
    }
    *out << ");" << endl;
  }  // end if

  *out << "END " << mname.latin1() << ";" << endl << endl;
}




/// Writes the 'architecture' block to the output stream
void ExportVHDL::writeArchitecture()
{
  using namespace std;

  QString bit_string;
  bool io_names_t, state_encoding_t, synchronous_reset_t;

  if (alliance)
  {
    io_names_t=false;
    state_encoding_t=false;
    synchronous_reset_t=true;
  }
  else
  {
    io_names_t=io_names;
    state_encoding_t=state_encoding;
    synchronous_reset_t=synchronous_reset;
  }


  if (use_std_logic)
    bit_string = "std_ulogic";
  else
    bit_string = "bit";

  *out << "ARCHITECTURE behave OF " << mname.latin1() << " IS" << endl << endl;
  if (!state_encoding_t)
  {
    QList<GState*> slist = machine->getSList();
    QString stmp;
    QMutableListIterator<GState*> it(slist);
    bool first=TRUE;
    for(; it.hasNext();)
    {
      GState* st = it.next();
      if(st->isDeleted())
        continue;
      if (!first)
	stmp += ", ";
      stmp += Utils::noWS(st->getStateName());
      first=FALSE;
    }
    *out << "TYPE state_type IS (" << stmp.latin1() << ");" << endl;
    *out << "SIGNAL next_state, current_state : state_type;" << endl << endl;
  }
  else
  {
    *out << "SIGNAL current_state, next_state : " << bit_string.latin1() << "_vector(" <<
      machine->getNumEncodingBits()-1
      << " DOWNTO 0);" << endl << endl; 
  }

  if (alliance)
  {
    *out << "-- comments for syf which comes with alliance : http://www-asim.lip6.fr/alliance"<<endl;
    *out << "-- interpreted as comments by other tools"<<endl;
    *out << "  --pragma CURRENT_STATE current_state" << endl;
    *out << "  --pragma NEXT_STATE next_state" << endl;
    *out << "  --pragma CLOCK clk" << endl << endl;
  }


  *out << "BEGIN" << endl;
  writeClockProcess();
  writeStateProcess();
  *out << "END behave;" << endl;
}




/// Writes the clock process to the output stream
void ExportVHDL::writeClockProcess()
{
  using namespace std;

  bool io_names_t, state_encoding_t, synchronous_reset_t;
  if (alliance)
  {
    io_names_t=false;
    state_encoding_t=false;
    synchronous_reset_t=true;
  }
  else
  {
    io_names_t=io_names;
    state_encoding_t=state_encoding;
    synchronous_reset_t=synchronous_reset;
  }

  if (alliance)
  {
    *out << "  state_register: PROCESS (clk)" << endl;
    *out << "  BEGIN" << endl;
  }
  else
  {
    if (synchronous_reset_t)
    {
      if (neg_reset)
	*out << "  state_register: PROCESS (srst_n, rst_n, clk)" << endl;
      else
	*out << "  state_register: PROCESS (srst_p, rst_p, clk)" << endl;
    }
    else
    {
      if (neg_reset)
	*out << "  state_register: PROCESS (rst_n, clk)" << endl;
      else
	*out << "  state_register: PROCESS (rst_p, clk)" << endl;
    }

    *out << "  BEGIN" << endl;

    if (neg_reset)
      *out << "    IF rst_n='0' THEN" << endl;
    else
      *out << "    IF rst_p='1' THEN" << endl;

    GState* stmp = machine->getInitialState();

    if (state_encoding_t)
    {
      *out << "      current_state <= \"";
      *out << Convert::intToBinStr(stmp->getEncoding(), machine->getNumEncodingBits()).latin1() 
	<< "\";" << endl;
    }
    else
    {
      *out << "      current_state <= " <<
	Utils::noWS(stmp->getStateName()).latin1() 
	<< ";" << endl;
    }

    if(use_std_logic)
      *out << "    ELSIF rising_edge(clk) THEN" << endl;
    else *out << "    ELSIF clock'event and clock='1' THEN" << endl;
    
    
  }

  if (synchronous_reset_t && !alliance)
  {
    GState* istate;
    istate = machine->getInitialState();
    
    if (neg_reset)
      *out << "      IF srst_n='0' THEN " << endl; 
    else
      *out << "      IF srst_p='1' THEN " << endl; 

    
    if (state_encoding_t)
    {
      *out << "      current_state <= \"";
      *out << Convert::intToBinStr(istate->getEncoding(), machine->getNumEncodingBits()).latin1() 
      << "\";" << endl;
    }
    else
    {
      *out << "        current_state <= " <<
	  Utils::noWS(istate->getStateName()).latin1() 
	  << ";" << endl;
    }
        
    *out << "      ELSE" << endl;
    *out << "        current_state <= next_state;" << endl;
    *out << "      END IF;" << endl;
  }
  else 
  {
    if (alliance)
      *out << "    IF ( clk = '1' AND NOT clk'STABLE ) THEN" << endl;
    *out << "      current_state <= next_state;" << endl;
  }

  *out << "    END IF;" << endl;
  *out << "  END PROCESS;" << endl << endl;
}



/// Writes the input/current_state process to the output stream
void ExportVHDL::writeStateProcess()
{
  using namespace std;

  GState* s;
  GTransition* t;
  QString tinfoi, tinfoo, sn;
  State* stmp;
  TransitionInfo* tinfo;
  IOInfo* iosingle;
  IOInfo* tioinfo;
  bool first, first_trans;
  Convert conv;
  QStringList inputs=machine->getInputNameList();
  QStringList outputs=machine->getOutputNameList();
  QStringList::const_iterator i;
  bool io_names_t, state_encoding_t, synchronous_reset_t;
  int cond_notation_t;

  if (machine->getType()==1) // Binary machine
    io_names_t=false;

  if (alliance && machine->getType()==0)
  {
    io_names_t=false;
    state_encoding_t=false;
    synchronous_reset_t=false;
    cond_notation_t=0; // IF-THEN notation
  }
  else
  {
    io_names_t=io_names;
    state_encoding_t=state_encoding;
    synchronous_reset_t=synchronous_reset;
    cond_notation_t=cond_notation;
  }
  

  *out << "  next_state_and_output_logic: PROCESS (current_state";
  if (io_names_t)
  {
    for (i=inputs.constBegin(); i != inputs.constEnd(); ++i)
      *out << ", " << i->latin1();
  }
  else
    *out << ", a";
  *out << ")\n";

  if (io_names_t)
  {
    // temporary variables
    if(machine->getNumInputs()>0)
      *out << "    VARIABLE temp_input : std_ulogic_vector("<< machine->getNumInputs()-1 << " DOWNTO 0);" << endl;
    if(machine->getNumMooreOutputs()>0)
      *out << "    VARIABLE temp_output : std_ulogic_vector("<< machine->getNumMooreOutputs()-1 << " DOWNTO 0);" << endl;
    if(machine->getNumOutputs()>0)
      *out << "    VARIABLE temp_mealy_output : std_ulogic_vector("<< machine->getNumOutputs()-1 << " DOWNTO 0);" << endl;
    // end temporary
  }

  *out << "  BEGIN" << endl;

  if (io_names_t)
  {
    i=inputs.constBegin();
    if(machine->getNumInputs() == 1)
      *out << "    temp_input(0) := " << i->latin1();
    else if(machine->getNumInputs() >1)
    {
      *out << "    temp_input := " << i->latin1();
      i++;
      for (; i != inputs.constEnd(); ++i)
	*out << " & " << i->latin1();
    }
    *out << ";\n";
  }

  if (alliance)
  {
    GState* istate;
    istate = machine->getInitialState();
    
    if (neg_reset)
      *out << "    IF srst_n='0' THEN " << endl; 
    else
      *out << "    IF srst_p='1' THEN " << endl; 

    
    if (state_encoding_t)
    {
      *out << "      next_state <= \"";
      *out << Convert::intToBinStr(istate->getEncoding(), machine->getNumEncodingBits()).latin1() 
      << "\";" << endl;
    }
    else
    {
      *out << "        next_state <= " <<
	  Utils::noWS(istate->getStateName()).latin1() 
	  << ";" << endl;
    }
        
    /*
    *out << "    IF ( rst_p = '1' ) THEN" << endl;
    if (state_encoding)
    {
      *out << "      next_state <= \"";
      *out << Convert::intToBinStr(istate->getEncoding(), machine->getNumEncodingBits()).latin1() 
      << "\";" << endl;
    }
    else
    {
      *out << "        next_state <= " <<
	  Utils::noWS(istate->getStateName()).latin1() 
	  << ";" << endl;
    }
    */
    *out << "    ELSE" << endl;
  }


  /*
  *out << "    -- default assignments" << endl;
  *out << "--    o <= \"";
  for(int i=0; i<machine->getNumOutputs(); i++)
    *out << "0";
  *out << "\";" << endl;
  *out << "--    next_state <= current_state;" << endl << endl;
  */

  *out << "    CASE current_state IS" << endl;

  QMutableListIterator<GState*> is(machine->getSList());

  for(; is.hasNext();)
  {
    s = is.next();
    sn = s->getStateName();
    sn.replace(QRegExp(" "), "_");
    if (s->countTransitions()>0)
    {
      if (state_encoding_t)
      {
	  *out << "      WHEN \"" <<
	    Convert::intToBinStr(s->getEncoding(),
	      machine->getNumEncodingBits()).latin1()
	    << "\" =>"; 
      }
      else
      {
	  *out << "      WHEN " << Utils::noWS(s->getStateName()).latin1() << " =>";
      }
      if (machine->getNumMooreOutputs() > 0) 
      {
	IOInfo* mout;
	QString smout;
	mout = s->getMooreOutputs();
	smout = mout->convertToBinStr();

	if (!smout.isEmpty())
	{
	  if (io_names_t)
	    *out << " temp_output := \"";
	  else
	    *out << " q <= \"";
	  QStringList outputs=machine->getMooreOutputList();
	  QStringList::const_iterator i=outputs.constBegin();
	  int c=0;
	  while (i != outputs.constEnd())
	  {
	    *out << smout[c].toLatin1();
	    c++;
	    ++i;
	  }
          *out << "\";\n";
	}
	else   
	  *out << endl;
      }
      else *out << "\n";
    }

    QMutableListIterator<GTransition*> it(s->tlist);
    
    if (cond_notation_t==1 && s->countTransitions()>0)
    {
      if (io_names_t)
	*out << "        CASE temp_input IS" << endl;
      else
	*out << "        CASE a IS" << endl;
    }

    first_trans=TRUE;

    for(; it.hasNext();)
    {
      t = it.next();
      tinfo = t->getInfo();
      tioinfo = tinfo->getInputInfo();

      if (!t->isDeleted() && t->getEnd())
      {
	IOInfoList iolist;
//	QList<IOInfo> iolist;
	bool inv = tioinfo->isInverted();
	if (cond_notation_t==0) // IF-THEN notation
	{
	  if (alliance)
	  {
	    // treat info as non inverted (nasty workaround)
	    tioinfo->setInvert(false);
	  }
	  tioinfo->convertToBinList(iolist, false);

	  if (alliance)
	    tioinfo->setInvert(inv);
	}
	else // CASE-WHEN notation
	  tioinfo->convertToBinList(iolist, true);
	
//	iolist.setAutoDelete(TRUE);
	
	QMutableListIterator<IOInfo*> ioit(iolist);

	if (cond_notation_t==0)
	{
	  if (first_trans)
	    *out << "        IF ";
	  else
	    *out << "        ELSIF ";

	  if (io_names_t)
	    *out << "temp_input";
	  else
	    *out << "a";

	  if (tioinfo->isInverted())
	    *out << "/=";
	  else
	    *out << "=";
	}
	else
	{
	  *out << "          WHEN ";
	}

	*out << "\"";

	first = TRUE;
	for(; ioit.hasNext();)
	{

	  iosingle = ioit.next();
	  tinfoi = iosingle->convertToBinStr();  
	  tinfoi.replace(QRegExp("x"), "-");

	  if (!tinfoi.isEmpty())
	  {
	    if (cond_notation_t==0) // IF-THEN notation
	    {
	      if (!first)
	      {
		if (tioinfo->isInverted())
		  *out << "and ";
		else
		  *out << "or ";
		if (io_names_t)
		  //*out << endl << "           temp_input";
		  *out << "temp_input";
		else
		  //*out << endl << "           a";
		  *out << "a";
		if (tioinfo->isInverted())
		  *out << "/=";
		else
		  *out << "=";
		*out << "\"";
	      }
	    }
	    else // CASE-WHEN notation
	    {
	      if (!first)
		*out << "| \"";
	    }

	    int slen = tinfoi.length();
	    int numin = machine->getNumInputs();
	    for(int k=slen; k<numin; k++)
	      *out << "0";

	    *out << tinfoi.latin1() << "\" ";
	    first=FALSE;
	  }
	}

	tinfoo = tinfo->getOutputsStrBin();
	if (cond_notation_t==0) // IF-THEN notation
	  *out << "THEN" << endl;
	else // CASE-WHEN notation
	  *out << "=>" << endl;

	if (!tinfoo.isEmpty())
	{
	  if (cond_notation_t==1)
	    *out << "  ";  // correct indent

	  if (io_names_t)
	    *out << "          temp_mealy_output := \"";
	  else
	    *out << "          o <= \"";

	  int slen = tinfoo.length();
	  int numout = machine->getNumOutputs();
	  for(int l=slen; l<numout; l++)
	    *out << "0";

	  *out << tinfoo.latin1() << "\";" << endl;
	}
	stmp = t->getEnd();
	if (stmp)
        {
	  sn = stmp->getStateName();
          sn.replace(QRegExp(" "), "_");
        }
	//if (stmp!=s)
	{
	  if (cond_notation_t==1)
	    *out << "  ";  // correct indent

	  if (state_encoding_t)
	  {
	    *out << "          next_state <= \"" 
	      << Convert::intToBinStr(stmp->getEncoding(),
		  machine->getNumEncodingBits()).latin1() << "\";" << endl;
	  }
	  else
	  {
	    *out << "          next_state <= " <<
	      Utils::noWS(stmp->getStateName()).latin1() << ";" << endl;
	  }
	}

	first_trans=FALSE;
      }
    }
    if (s->countTransitions()>0)
    {
      if (cond_notation_t==0) // IF-THEN notation
      {
	if (!alliance)
	{
	  *out << "        ELSE" << endl;
	  *out << "          next_state <= current_state;" << endl;
	}
	*out << "        END IF;" << endl;
      }
      else // CASE-WHEN notation
	*out << "        END CASE;" << endl;   // temp_input
    }
  }

  *out << "      WHEN OTHERS => null;" << endl;
  *out << "    END CASE;" << endl;
  if (alliance)
    *out << "    END IF;" << endl;

  if (io_names_t)
  {
    if(machine->getNumMooreOutputs()>0)
    {
      QStringList outputs=machine->getMooreOutputList();
      QStringList::const_iterator i=outputs.constBegin();
      int c=machine->getNumMooreOutputs()-1;
      while (i != outputs.constEnd())
      {
	*out << "    " << i->latin1() << " <= temp_output(" << c << ");\n";
	c--;
	++i;
      }
    }
    if(machine->getNumOutputs()>0)
    {
      QStringList outputs=machine->getOutputNameList();
      QStringList::const_iterator i=outputs.constBegin();
      int c=machine->getNumOutputs()-1;
      while (i != outputs.constEnd())
      {
	*out << "    " << i->latin1() << " <= temp_mealy_output(" << c << ");\n";
	c--;
	++i;
      }
    }
  }

  *out << "  END PROCESS;" << endl << endl;
}


/// Writes the machine description to the output stream
void ExportVHDL::writeDescription()
{
  using namespace std;


  *out << endl;
  if(machine->getNumInputs()>0)
  {
    *out << "-- Inputs:";
    QStringList inputs=machine->getInputNameList();
    QStringList::const_iterator i;
    for(i=inputs.constBegin(); i!=inputs.constEnd(); ++i)
    {
      *out << "   "<< i->latin1();
    }
    *out << endl;
  }
  if(machine->getNumOutputs()>0)
  {
    *out << "-- Mealy Outputs:";
    QStringList outputs=machine->getOutputNameList();
    QStringList::const_iterator i;
    for(i=outputs.constBegin(); i!=outputs.constEnd(); ++i)
    {
      *out << "   "<< i->latin1();
    }
    *out << endl;
  }


  QList<GState*>states= machine->getSList();
  QList<GState*>::const_iterator is;
  int *textLen=new int[machine->getNumMooreOutputs()+1];
  textLen[0]=12;

  for(is=states.constBegin(); is!=states.constEnd(); ++is)
  {
    if(!(*is)->isDeleted() && (*is)->getStateName().length()>textLen[0])
      textLen[0]=(*is)->getStateName().length();
  }
  textLen[0]++;

  *out << "-- State/Output" << QString(textLen[0]-13,' ').latin1();


  if(machine->getNumMooreOutputs()>0)
  {
    int spc=1;
    QStringList outputs=machine->getMooreOutputList();
    QStringList::const_iterator i;
    for(i=outputs.constBegin(); i!=outputs.constEnd(); ++i)
    {
      textLen[spc]=i->length()+1;
      spc++;
      *out << " "<< i->latin1();
    }
  }
  *out << endl;

  IOInfo*mooreOutput;
  QString sMooreOutput;

  for(is=states.constBegin(); is!=states.constEnd(); ++is)
  {
    if((*is)->isDeleted())
      continue;
    int spc=0;
    *out << "-- " << (*is)->getStateName().latin1() << QString(textLen[spc]-(*is)->getStateName().length(),' ').latin1();
    spc++;
    if(machine->getNumMooreOutputs()>0)
    {
      mooreOutput=(*is)->getMooreOutputs();
      sMooreOutput=mooreOutput->convertToBinStr();
      int c=0;
      while(c<sMooreOutput.length())
      {
        *out << sMooreOutput[c].toLatin1() << QString(textLen[spc]-1,' ').latin1();
        c++;
        spc++;
      }
    }
    *out << endl;
  }
  *out << endl;
  delete[] textLen;
}

#ifdef FHA
/// Writes a comment to the top of the output stream 
void ExportVHDL::writeHeader(QString commentstart, QString commentend)
{
  *out << commentstart.latin1() << "----------------------------------------------------------------" << commentend.latin1() << "\n";
  *out << commentstart.latin1() << " Module     : " << machine->getName().latin1() << commentend.latin1() << "\n";
  *out << commentstart.latin1() << "----------------------------------------------------------------" << commentend.latin1() << "\n";
  *out << commentstart.latin1() << " Author     : " << commentend.latin1() << "\n";
  *out << commentstart.latin1() << " Company    : University of Applied Sciences Augsburg" << commentend.latin1() << "\n";
  *out << commentstart.latin1() << "----------------------------------------------------------------" << commentend.latin1() << "\n";
  *out << commentstart.latin1() << " Description: " << commentend.latin1() << "\n";
  *out << commentstart.latin1() << "----------------------------------------------------------------" << commentend.latin1() << "\n";
  *out << commentstart.latin1() << " Revisions : " << machine->getVersion().latin1() << " - " << commentend.latin1() << "\n";
  *out << commentstart.latin1() << "----------------------------------------------------------------" << commentend.latin1() << "\n";
}
#endif


