/*
Copyright (C) 2000,2001 Stefan Duffner 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <qtimer.h>
//Added by qt3to4:
#include <QCloseEvent>
#include <QPixmap>
#include "SimulatorDlgImpl.h"
#include "MainWindow.h"
#include "Project.h"
#include "Convert.h"
#include "Machine.h"
#include "Const.h"
#include "TransitionInfo.h"
#include "../pics/redlight.xpm"
#include "../pics/greenlight.xpm"
#include "../pics/greylight.xpm"

/** 
 *  Constructs a SimulatorDlgImpl which is a child of 'parent', with the 
 *  name 'name' and widget flags set to 'f' 
 *
 *  The dialog will by default be modeless, unless you set 'modal' to
 *  TRUE to construct a modal dialog.
 */
SimulatorDlgImpl::SimulatorDlgImpl( QWidget* parent,  const char* name, bool modal, Qt::WFlags fl )
    : SimulatorDlg( parent, name, modal, fl )
{
    main = (MainWindow*)parent;
    timer = new QTimer(this);

    redlight = new QPixmap((const char**)redlight_xpm);
    greenlight = new QPixmap((const char**)greenlight_xpm);
    greylight = new QPixmap((const char**)greylight_xpm);

    connect(timer, SIGNAL(timeout()), this, SIGNAL(inputSent()));
}

/**  
 *  Destroys the object and frees any allocated resources
 */
SimulatorDlgImpl::~SimulatorDlgImpl()
{
  delete timer;
  delete redlight;
  delete greenlight;
  delete greylight;
}


/// Initialises the dialog
void SimulatorDlgImpl::init()
{
  resetButtons();

  if (isIBinChecked())
    le_inputs->setMaxLength(main->project->machine->getNumInputs());
  else if (isIHexChecked())
    le_inputs->setMaxLength(int((double)main->project->machine->getNumInputs()/4+0.9));
  else 
    le_inputs->setMaxLength(int((double)main->project->machine->getNumInputs()/8 +0.9)*4);

  setInputs("");
  le_inputs->setFocus();
  setState(Off);
}

/**
 * Close the simulator.
 * Emits the closing signal and hides the dialog. Is invoked when the users
 * clicks the Close button.
 */
void SimulatorDlgImpl::closeSim()
{
  emit closing();
  hide();
}

/**
 * Is invoked when the user closes the window.
 * Emits the closing signal and hides the dialog.
 */
void SimulatorDlgImpl::closeEvent(QCloseEvent* e)
{
  emit closing();
  e->accept();
}

/**
 * User has clicked the Send button.
 */
void SimulatorDlgImpl::send()
{
  emit inputSent();
}

/**
 * User has clicked the Reset button.
 */
void SimulatorDlgImpl::reset()
{
  emit resetPressed();
}

/// Stops the timer for the clock.
void SimulatorDlgImpl::stopTimer()
{
  timer->stop();
}

/// Selects the contents of the 'inputs' field and sets the focus
void SimulatorDlgImpl::selectFirst()
{
  le_inputs->selectAll();
  le_inputs->setFocus();
}


/**
 * Is invoked when the text in the inputs field changes.
 * The buttons for the bits 0-15 are changed according to the string @a in.
 */
void SimulatorDlgImpl::inputsChanged(const QString& in)
{
  Convert conv;
  int len;
  char* bin;
//  char* bininv;
  IOInfoBin io(IO_MealyIn), ioinv(IO_MealyIn);

  len = main->project->machine->getNumInputs();
//  bin = new char[len+1];
//  bininv = new char[len+1];
  if (isIBinChecked())
    ioinv = conv.binStrToX10(len, in, IO_MealyIn);
  else if (isIHexChecked())
    ioinv = conv.hexStrToX10(len, in, IO_MealyIn);
  else 
  {
    unsigned char ascii[MAX_CHARARRAY_LENGTH];
    int alen;

    conv.resolveEscapes(in, ascii, MAX_CHARARRAY_LENGTH, alen);
    ioinv = conv.asciiToX10(len, ascii, alen, IO_MealyIn);
  }

//  conv.invertX10(len, bininv, bin);
  io = ioinv;
  io.invertBits();
  bin = io.getInfo();

  if (pb_bin0->isEnabled())
    pb_bin0->setOn(bin[0]==1);
  if (pb_bin1->isEnabled())
    pb_bin1->setOn(bin[1]==1);
  if (pb_bin2->isEnabled())
    pb_bin2->setOn(bin[2]==1);
  if (pb_bin3->isEnabled())
    pb_bin3->setOn(bin[3]==1);
  if (pb_bin4->isEnabled())
    pb_bin4->setOn(bin[4]==1);
  if (pb_bin5->isEnabled())
    pb_bin5->setOn(bin[5]==1);
  if (pb_bin6->isEnabled())
    pb_bin6->setOn(bin[6]==1);
  if (pb_bin7->isEnabled())
    pb_bin7->setOn(bin[7]==1);
  if (pb_bin8->isEnabled())
    pb_bin8->setOn(bin[8]==1);
  if (pb_bin9->isEnabled())
    pb_bin9->setOn(bin[9]==1);
  if (pb_bin10->isEnabled())
    pb_bin10->setOn(bin[10]==1);
  if (pb_bin11->isEnabled())
    pb_bin11->setOn(bin[11]==1);
  if (pb_bin12->isEnabled())
    pb_bin12->setOn(bin[12]==1);
  if (pb_bin13->isEnabled())
    pb_bin13->setOn(bin[13]==1);
  if (pb_bin14->isEnabled())
    pb_bin14->setOn(bin[14]==1);
  if (pb_bin15->isEnabled())
    pb_bin15->setOn(bin[15]==1);

//  delete [] bin;
//  delete [] bininv;
}

/// Called when return is pressed in the 'inputs' field
void SimulatorDlgImpl::returnPressedInputs()
{
  /*
  if (pb_send->isEnabled())
    send();
    */
}


/// Button 0 was pressed
void SimulatorDlgImpl::button0Pressed()
{
  updateInputsStr();
}

/// Button 1 was pressed
void SimulatorDlgImpl::button1Pressed()
{
  updateInputsStr();
}

/// Button 2 was pressed
void SimulatorDlgImpl::button2Pressed()
{
  updateInputsStr();
}

/// Button 3 was pressed
void SimulatorDlgImpl::button3Pressed()
{
  updateInputsStr();
}

/// Button 4 was pressed
void SimulatorDlgImpl::button4Pressed()
{
  updateInputsStr();
}

/// Button 5 was pressed
void SimulatorDlgImpl::button5Pressed()
{
  updateInputsStr();
}

/// Button 6 was pressed
void SimulatorDlgImpl::button6Pressed()
{
  updateInputsStr();
}

/// Button 7 was pressed
void SimulatorDlgImpl::button7Pressed()
{
  updateInputsStr();
}

/// Button 8 was pressed
void SimulatorDlgImpl::button8Pressed()
{
  updateInputsStr();
}

/// Button 9 was pressed
void SimulatorDlgImpl::button9Pressed()
{
  updateInputsStr();
}

/// Button 10 was pressed
void SimulatorDlgImpl::button10Pressed()
{
  updateInputsStr();
}

/// Button 11 was pressed
void SimulatorDlgImpl::button11Pressed()
{
  updateInputsStr();
}

/// Button 12 was pressed
void SimulatorDlgImpl::button12Pressed()
{
  updateInputsStr();
}

/// Button 13 was pressed
void SimulatorDlgImpl::button13Pressed()
{
  updateInputsStr();
}

/// Button 14 was pressed
void SimulatorDlgImpl::button14Pressed()
{
  updateInputsStr();
}

/// Button 15 was pressed
void SimulatorDlgImpl::button15Pressed()
{
  updateInputsStr();
}


/**
 * Updates the string in the inputs field according to the state of the buttons.
 */
void SimulatorDlgImpl::updateInputsStr()
{
  int len;
  QString s;
  Convert conv;
  char* bin;
  IOInfoBin io(IO_MealyIn), ioinv(IO_MealyIn);

  len = main->project->machine->getNumInputs();
  s = le_inputs->text();

  if (isIBinChecked())
    ioinv = conv.binStrToX10(len, s, IO_MealyIn);
  else if (isIHexChecked())
    ioinv = conv.hexStrToX10(len, s, IO_MealyIn);
  else 
  {
    unsigned char ascii[MAX_CHARARRAY_LENGTH];
    int alen;

    conv.resolveEscapes(s, ascii, MAX_CHARARRAY_LENGTH, alen);
    ioinv = conv.asciiToX10(len, ascii, alen, IO_MealyIn);
  }

  io = ioinv;
  io.invertBits();
  bin = io.getInfo();

  if (len>0)
    bin[0]=pb_bin0->isOn();
  if (len>1)
    bin[1]=pb_bin1->isOn();
  if (len>2)
    bin[2]=pb_bin2->isOn();
  if (len>3)
    bin[3]=pb_bin3->isOn();
  if (len>4)
    bin[4]=pb_bin4->isOn();
  if (len>5)
    bin[5]=pb_bin5->isOn();
  if (len>6)
    bin[6]=pb_bin6->isOn();
  if (len>7)
    bin[7]=pb_bin7->isOn();
  if (len>8)
    bin[8]=pb_bin8->isOn();
  if (len>9)
    bin[9]=pb_bin9->isOn();
  if (len>10)
    bin[10]=pb_bin10->isOn();
  if (len>11)
    bin[11]=pb_bin11->isOn();
  if (len>12)
    bin[12]=pb_bin12->isOn();
  if (len>13)
    bin[13]=pb_bin13->isOn();
  if (len>14)
    bin[14]=pb_bin14->isOn();
  if (len>15)
    bin[15]=pb_bin15->isOn();
    
  bin[len]=10;

//  io.setInfo(bin);
  ioinv = io;
  ioinv.invertBits();

  if (isIBinChecked())
    s = conv.X10ToBinStr(ioinv);
  else if (isIHexChecked())
    s = conv.X10ToHexStr(ioinv);
  else
  {
    unsigned char ascii[MAX_CHARARRAY_LENGTH];
    int alen;

    if (!conv.X10ToASCII(ioinv, ascii, MAX_CHARARRAY_LENGTH, alen))
      s="";
    else
      s=conv.asciiToReadableStr(ascii, alen);
  }

  le_inputs->setText(s);

}


/// Called when the 'Binary' button for the inputs is pressed.
void SimulatorDlgImpl::iBinPressed()
{
  QString tmp;
  Convert conv;

  if (isIHexChecked())
  {
    tmp = conv.hexStrToBinStr(main->project->machine->getNumInputs(), le_inputs->text());
//    le_inputs->setText(tmp);
//    rb_ibin->setChecked(TRUE);
//    inputsChanged(tmp);
  }
  else if (isIASCIIChecked())
  {
    unsigned char ascii[MAX_CHARARRAY_LENGTH];
    int alen;

    conv.resolveEscapes(le_inputs->text(), ascii, MAX_CHARARRAY_LENGTH, alen);
    tmp = conv.asciiToBinStr(main->project->machine->getNumInputs(), ascii, alen);
//    le_inputs->setText(tmp);
//    rb_ibin->setChecked(TRUE);
//    inputsChanged(tmp);
  }
  le_inputs->setMaxLength(main->project->machine->getNumInputs());
  le_inputs->setText(tmp);
  rb_ibin->setChecked(TRUE);
  inputsChanged(tmp);
}


/// Called when the 'Hexadecimal' button for the inputs is pressed.
void SimulatorDlgImpl::iHexPressed()
{
  QString tmp;
  Convert conv;
  int maxlen;

  if (isIBinChecked())
  {
    tmp = conv.binStrToHexStr(le_inputs->text());
//    le_inputs->setText(tmp);
//    rb_ihex->setChecked(TRUE);
//    inputsChanged(tmp);
  }
  else if (isIASCIIChecked())
  {
    unsigned char ascii[MAX_CHARARRAY_LENGTH];
    int alen;

    conv.resolveEscapes(le_inputs->text(), ascii, MAX_CHARARRAY_LENGTH, alen);
    tmp = conv.asciiToHexStr(ascii, alen);
//    le_inputs->setText(tmp);
//    rb_ihex->setChecked(TRUE);
//    inputsChanged(tmp);
  }
  maxlen = int((double)main->project->machine->getNumInputs()/4+0.9);
  tmp = tmp.right(maxlen);
  le_inputs->setMaxLength(maxlen);
  le_inputs->setText(tmp);
  rb_ihex->setChecked(TRUE);
  inputsChanged(tmp);
}


/// Called when the 'ASCII' button for the inputs is pressed
void SimulatorDlgImpl::iASCIIPressed()
{
  QString tmp;
  Convert conv;

  if (isIBinChecked())
  {
    unsigned char ascii[MAX_CHARARRAY_LENGTH];
    int alen;

    conv.binStrToASCII(le_inputs->text(), ascii, MAX_CHARARRAY_LENGTH, alen);
    tmp = conv.asciiToReadableStr(ascii, alen);
//    le_inputs->setText(tmp);
//    rb_iascii->setChecked(TRUE);
//    inputsChanged(tmp);
  }
  else if (isIHexChecked())
  {
    unsigned char ascii[MAX_CHARARRAY_LENGTH];
    unsigned char single;
    int alen;

    conv.hexStrToASCII(le_inputs->text(), ascii, MAX_CHARARRAY_LENGTH, alen);
    if (alen>0)
    {
      single = ascii[alen-1];
      tmp = conv.asciiToReadableStr(&single, 1);
    }
//    le_inputs->setText(tmp);
//    rb_iascii->setChecked(TRUE);
//    inputsChanged(tmp);
  }
  le_inputs->setMaxLength(int((double)main->project->machine->getNumInputs()/8 +0.9)*4);
  le_inputs->setText(tmp);
  rb_iascii->setChecked(TRUE);
  inputsChanged(tmp);
}

/// Called when the 'Binary' button for the outputs is pressed.
void SimulatorDlgImpl::oBinPressed()
{
  QString tmp;
  Convert conv;

  if (isOHexChecked())
  {
    tmp = conv.hexStrToBinStr(main->project->machine->getNumOutputs(), lb_outputs->text());
    lb_outputs->setText(tmp);
  }
  else if (isOASCIIChecked())
  {
    unsigned char ascii[MAX_CHARARRAY_LENGTH];
    int alen;
    conv.resolveEscapes(lb_outputs->text(), ascii, MAX_CHARARRAY_LENGTH, alen);
    tmp = conv.asciiToBinStr(main->project->machine->getNumOutputs(), ascii, alen);
    lb_outputs->setText(tmp);
  }
}


/// Called when the 'Hexadecimal' button for the outputs is pressed.
void SimulatorDlgImpl::oHexPressed()
{
  QString tmp;
  Convert conv;

  if (isOBinChecked())
  {
    tmp = conv.binStrToHexStr(lb_outputs->text());
    lb_outputs->setText(tmp);
  }
  else if (isOASCIIChecked())
  {
    unsigned char ascii[MAX_CHARARRAY_LENGTH];
    int alen;
    conv.resolveEscapes(lb_outputs->text(), ascii, MAX_CHARARRAY_LENGTH, alen);
    tmp = conv.asciiToHexStr(ascii, alen);
    lb_outputs->setText(tmp);
  }
}

/// Called when the 'ASCII' button for the outputs is pressed.
void SimulatorDlgImpl::oASCIIPressed()
{
  QString tmp;
  Convert conv;

  if (isOBinChecked())
  {
    unsigned char ascii[MAX_CHARARRAY_LENGTH];
    int alen;

    conv.binStrToASCII(lb_outputs->text(), ascii, MAX_CHARARRAY_LENGTH, alen, FALSE);
    tmp = conv.asciiToReadableStr(ascii, alen);
    lb_outputs->setText(tmp);
  }
  else if (isOHexChecked())
  {
    unsigned char ascii[MAX_CHARARRAY_LENGTH];
    int alen;

    conv.hexStrToASCII(lb_outputs->text(), ascii, MAX_CHARARRAY_LENGTH, alen, FALSE);
    tmp = conv.asciiToReadableStr(ascii, alen);
    lb_outputs->setText(tmp);
  }
}



/// Sets all button states to Off
void SimulatorDlgImpl::resetButtons()
{
  if (main->project->machine)
  {
    if (main->project->machine->getType()==Binary)
    {
      rb_ibin->setChecked(TRUE);
      rb_obin->setChecked(TRUE);
    }
    else
    {
      rb_iascii->setChecked(TRUE);
      rb_oascii->setChecked(TRUE);
    }
      
  }

  pb_bin0->setOn(FALSE);
  pb_bin1->setOn(FALSE);
  pb_bin2->setOn(FALSE);
  pb_bin3->setOn(FALSE);
  pb_bin4->setOn(FALSE);
  pb_bin5->setOn(FALSE);
  pb_bin6->setOn(FALSE);
  pb_bin7->setOn(FALSE);
  pb_bin8->setOn(FALSE);
  pb_bin9->setOn(FALSE);
  pb_bin10->setOn(FALSE);
  pb_bin11->setOn(FALSE);
  pb_bin12->setOn(FALSE);
  pb_bin13->setOn(FALSE);
  pb_bin14->setOn(FALSE);
  pb_bin15->setOn(FALSE);
  pb_clock->setOn(FALSE);
}


/// Called when the 'clock' button is pressed
void SimulatorDlgImpl::clockPressed()
{
  double hz;
  double msec;
  hz = frequency->value();
//  hz /= 10;
  msec = 1000/hz;

  if (!pb_clock->isOn())
  {
    timer->start((int)msec);
    lb_frequency->setEnabled(FALSE);
    frequency->setEnabled(FALSE);
    pb_send->setEnabled(FALSE);
  }
  else if (timer->isActive())
  {
    timer->stop();
    lb_frequency->setEnabled(TRUE);
    frequency->setEnabled(TRUE);
    pb_send->setEnabled(TRUE);
  }
}


/**
 * Sets the state of the machine to @a s
 * @param s State of the machine
 *    - Busy: Machine hasn't reached an end state (red light)
 *    - Finished: Machine has reached an end state (green light)
 *    - Otherwise undefined (grey light)
 */
void SimulatorDlgImpl::setState(int s)
{
  switch (s)
  {
    case Busy:     pm_end->setPixmap(*redlight); break;
    case Finished: pm_end->setPixmap(*greenlight); break;
    default: pm_end->setPixmap(*greylight); break;
  }
}

